
"""Utilities to support Mixminion usage."""

import re, os.path, time, calendar
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from twisted.python import failure
from twisted.internet import protocol, defer, reactor
from twisted.application import service, internet
from petmail import util

class HelperPP(protocol.ProcessProtocol):
    debug = 0
    def __init__(self, inputs, outputs=[]):
        # inputs is a dict which maps fd numbers to strings
        self.inputs = inputs
        # outputs is a list of fds which should be read
        self.outputs = {}
        for fd in outputs:
            self.outputs[fd] = "" 
        # this Deferred will be fired (with the outputs dict) when the
        # process is complete
        self.deferred = defer.Deferred()

    def connectionMade(self):
        if self.debug: print "connectionMade"
        for fd, text in self.inputs.items():
            self.transport.writeToChild(fd, text)
            self.transport.closeChildFD(fd)

    def childDataReceived(self, childFD, data):
        if self.debug: print "childDataReceived", data
        self.outputs[childFD] += data

    def processEnded(self, status):
        if self.debug: print "processEnded"
        if status.value.signal:
            self.deferred.errback((status, self.outputs))
        else:
            self.rc = status.value.exitCode
            if self.rc:
                self.deferred.errback((status, self.outputs))
            else:
                self.deferred.callback(self.outputs)

class VersionGetter(object):
    debug = 0
    def get(self, mixminion_name):
        mixminion_name = os.path.expanduser(mixminion_name)
        cmd = (mixminion_name, "version")
        if self.debug: print "cmd:",cmd
        pp = HelperPP({}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._done)
        d.addErrback(self._error)
        return d

    def _done(self, outputs):
        if self.debug: print "VersionGetter.done"
        stdout = outputs[1]
        r = re.search(r'Mixminion version (\S+)$', stdout, re.M)
        if not r:
            return (None, stdout)
        return (r.group(1), None)

    def _error(self, why):
        if self.debug: print "VersionGetter._error"
        if isinstance(why, failure.Failure):
            return (None, why)
        else:
            status, outputs = results
            return (None, outputs)


def getVersion(mixminion_name="mixminion"):
    """I attempt to run mixminion and determine what version is available. By
    default I use a bare 'mixminion' command (which means your $PATH will be
    searched), but you can override this to use /usr/local/bin/mixminion or
    ~/bin/mixminion or whatever.

    I return a Deferred which fires with a tuple, either (version, None),
    or (None, errortext)."""

    return VersionGetter().get(mixminion_name)

class Decoder(object):
    def decode_plaintext(self, encoded, mm_path, rcfile):
        cmd = (mm_path, "decode", "--input", "-")
        if rcfile:
            cmd = cmd + ("--config", rcfile)
        pp = HelperPP({0:encoded}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._done)
        return d

    def _done(self, outputs):
        stdout = outputs[1]
        return stdout

def decode_plaintext(encoded, mm_path="mixminion", rcfile=None):
    return Decoder().decode_plaintext(encoded, mm_path, rcfile)

class SURB(object):
    """I represent a single SURB.
    """
    # TODO: make this smaller, there could be thousands of them
    granted = False
    redeemed = False
    expired = False

    def __init__(self, identity, expires):
        self.identity = identity
        self.expires = expires
        self.watchers = []

    def setSURBid(self, surbid):
        self.surbid = surbid
    def setReplyBlock(self, text):
        self.text = text

    def get(self):
        """Retrieve the text of the SURB, suitable for giving to someone
        else."""
        assert not self.granted
        self.granted = True
        return self.text

    def redeem(self):
        assert self.granted
        assert not self.redeemed
        self.redeemed = True

    def notifyOnExpiration(self, watcher):
        self.watchers.append(watcher)
    def dontNotifyOnExpiration(self, watcher):
        self.watchers.remove(watcher)

    def expire(self):
        assert not self.redeemed
        assert not self.expired
        self.expired = True
        for w in self.watchers:
            w(self)

class MixminionContext(service.MultiService):
    """I maintain a Mixminion context on behalf of an Agent.
    This context has its own .mixminionrc and a separate home directory.

    I keep track of all SURBs created in this context, from minting to final
    redemption or expiration.
    """

    startedTimer = False
    pollinterval = 3600 # seconds

    def __init__(self, basedir):
        service.MultiService.__init__(self)
        """Create the context. 'basedir' should be a subdirectory of the
        parent agent's base directory.
        """
        self.basedir = basedir
        if not os.path.isdir(basedir):
            os.mkdir(basedir, 0700)
        self.mmrcfile = os.path.join(basedir, "mixminion.rc")
        self.mmrcfile = os.path.abspath(self.mmrcfile)
        f = open(self.mmrcfile, "w")
        f.write("""
[Host]

[DirectoryServers]
DirectoryTimeout: 1 minute

[User]
UserDir: %s

[Security]

[Network]
ConnectionTimeout: 60 seconds

        """ % self.basedir)
        self.surbids = {} # keys are surbids, values are SURBs
        self.outstanding = []
        self.passphrase = "Sekrit" # TODO: generate randomly

    def set_mm_path(self, mm_path):
        self.mm_path = mm_path

    def startService(self):
        if not self.startedTimer:
            # note that the TimerService is not pickleable until it has been
            # stopped.
            # TODO: consider using the deferred to prevent overlapping checks
            # TODO: SURB expiration is quantized to midnight GMT, so this
            #  could be done less frequently. Remember that the SURB could
            #  be in flight when it expires.
            timer = internet.TimerService(self.pollinterval, self.check)
            timer.setServiceParent(self)
            self.startedTimer = True
        service.MultiService.startService(self)
        self.check()

    def remove(self, s):
        self.outstanding.remove(s)
        if self.surbids.get(s.surbid):
            del self.surbids[s.surbid]

    def expire(self, s):
        self.remove(s)
        s.expire()

    def redeem(self, surbid):
        """I should be called when a message arrives that used the given
        SURB. I mark the SURB as having been redeemed."""
        s = self.surbids.get(surbid)
        if s:
            self.remove(s)
            s.redeem()

    def check(self):
        # TODO: sort .outstanding by expiration date to make this faster
        now = util.now()
        expired = []
        for s in self.outstanding:
            if s.expires and s.expires > now:
                expired.append(s)
        for s in expired:
            self.expire(s)

    def checkDirectoryDownload(self):
        """Utility method to see if mixminion can reach its directory
        servers. It needs to do this the first time it is run, and every
        time it thinks its list is out of date.

        This method returns a Deferred which fires back with True or False.
        If false, mixminion cannot be used (probably because we are not
        currently online). """
        
        cmd = (self.mm_path, "list-servers",
               "--config=%s" % self.mmrcfile,
               "-D", "yes", # force download
               )
        pp = HelperPP({}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={1:"r", 2:"r"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._checkDD_done)
        d.addErrback(self._checkDD_failed)
        return d

    def _checkDD_done(self, outputs):
        #print outputs[2]
        return True
    def _checkDD_failed(self, why):
        return False

    def mint(self, dest, idstring, lifetime=7):
        """Create a new SURB for the given identity-string. Each
        identity-string requires mixminion to maintain a separate secret
        key, so don't go nuts with them. These strings are strictly local,
        so it is safe to use the Owner's ID number for them.
        """

        # per the protocol, the SURB will expire 'lifetime' days after the
        # previous GMT midnight. This routine taken from mixminion.Common
        when = util.now() + 24*60*60*lifetime
        yyyy,MM,dd = time.gmtime(when)[0:3]
        expires = calendar.timegm((yyyy,MM,dd,0,0,0,0,0,0))

        surb = SURB(idstring, expires)

        cmd = (self.mm_path, "generate-surb",
               "--config=%s" % self.mmrcfile,
               "--to=%s" % dest,
               "--identity=%s" % idstring,
               "--lifetime=%d" % lifetime,
               "--passphrase-fd=%d" % 3,
               )
        pp = HelperPP({3:self.passphrase}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={1:"r", 2:"r", 3:"w"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._mint_done, surb)
        return d

    def _mint_done(self, outputs, surb):
        stdout = outputs[1]
        #print "stderr", outputs[2]
        surbid = "unknown" # TODO: need MM --status-fd API to get this
        surb.setReplyBlock(stdout)
        surb.setSURBid(surbid)

        self.outstanding.append(surb) # TODO: sort by expiration date
        self.surbids[surbid] = surb
        return surb

    def decrypt(self, message):
        """Decrypt a mixminion message. Returns a Deferred which fires with
        a tuple (plaintext, identity, surbid). If the message was encrypted
        with a SURB, the SURB's 'identity' string and surbid are returned,
        and the SURB is marked as having been used. If the message was not
        encrypted (message-type: plaintext), 'identity' and 'surbid' will be
        None."""

        cmd = (self.mm_path, "decode",
               "--config=%s" % self.mmrcfile,
               "--verbose", # want INFO lines
               "--input=-",
               "--passphrase-fd=%d" % 3,
               )
        pp = HelperPP({0:message, 3:self.passphrase}, [1,2])
        d = pp.deferred
        try:
            p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                     childFDs={0:"w", 1:"r", 2:"r", 3:"w"})
        except:
            print "oops"
            return defer.succeed((None, failure.Failure()))
        d.addCallback(self._decrypt_done)
        return d

    def _decrypt_done(self, outputs):
        stdout = outputs[1]
        stderr = StringIO(outputs[2])
        surbid, idstring = None, None
        for line in stderr.readlines():
            r = re.search(r"[INFO] Decoded reply message to identity '(\w*)'",
                          line)
            if r:
                surbid = "unknown" # TODO: need MM --status-fd API to get this
                idstring = r.group(1)
        if surbid:
            self.redeem(surbid)
        return (stdout, idstring, surbid)




def test():
    d = getVersion()
    def done(results):
        version, error = results
        if version:
            print "good!:", version
        else:
            print "bad!:", error
        reactor.stop()
        # this could be called before reactor.run() is
    def oops(why):
        print why
        reactor.stop()
    d.addCallback(done)
    d.addErrback(oops)
    print "running"
    reactor.run()

if __name__ == '__main__':
    test()
