
from petmail import log
from twisted.trial import unittest
import time

class LogTest(unittest.TestCase):
    def testBasics(self):
        l = log.Log("foo", maxItems=3)
        l.add("one")
        l.add("two")
        l.add("three")
        self.failUnlessEqual(len(l.getItems()), 3)
        l.add("four") # should displace "one"
        items = l.getItems()
        self.failUnlessEqual(len(items), 3)
        self.failUnlessEqual(items[0], l.getItem(0))
        m0 = items[0].getMessage()
        self.failUnlessEqual(type(m0), tuple)
        self.failUnlessEqual(m0[0], "two")
        self.failUnlessEqual(l.getMessage(0)[0], "two")
        self.failUnlessEqual(l.getMessage(-1)[0], "four")

    def testOnePriority(self):
        l = log.Log("foo", numClasses=1, maxItems=3)
        l.add("one", pclass=0)
        l.add("two", pclass=0)
        l.add("three", pclass=0)
        l.add("four", pclass=0)
        items = l.getItems()
        self.failUnlessEqual(len(items), 3)
        self.failUnlessEqual(l.getMessage(0), ("two",))
        self.failUnlessEqual(l.getMessage(-1), ("four",))

    def testPriorities(self):
        l = log.Log("foo", numClasses=3, maxItems=3)
        l.add("a", pclass=0) # 0a
        self.failUnlessEqual(len(l.getItems(None)), 1)
        self.failUnlessEqual(len(l.getItems(0)), 1)
        self.failUnlessEqual(len(l.getItems(1)), 0)
        self.failUnlessEqual(l.getMessage(0, None), ("a",))
        self.failUnlessEqual(l.getMessage(0, 0), ("a",))
        self.failUnlessRaises(IndexError, l.getMessage, 0, 1)

        l.add("b", pclass=0) # 0a 0b
        l.add("c", pclass=1) # 0a 0b 1c
        l.add("d", pclass=2) # 0a 0b 1c 2d
        l.add("e", pclass=0) # 0a 0b 1c 2d 0e
        l.add("f", pclass=1) # 0a 0b 1c 2d 0e 1f
        self.failUnlessEqual(len(l.getItems(None)), 6)
        self.failUnlessEqual(len(l.getItems(0)), 3)
        self.failUnlessEqual(len(l.getItems(1)), 2)
        self.failUnlessEqual(len(l.getItems(2)), 1)
        self.failUnlessEqual(l.getMessage(-1, None), ("f",))
        self.failUnlessEqual(l.getMessage(-1, 0), ("e",))
        self.failUnlessEqual(l.getMessage(-1, 1), ("f",))
        self.failUnlessEqual(l.getMessage(-1, 2), ("d",))
        self.failUnlessEqual(l.getMessage(0, None), ("a",))
        self.failUnlessEqual(l.getMessage(0, 0), ("a",))
        self.failUnlessEqual(l.getMessage(0, 1), ("c",))
        self.failUnlessEqual(l.getMessage(0, 2), ("d",))

        l.add("g", pclass=0) # 0b 1c 2d 0e 1f 0g
        l.add("h", pclass=2) # 0b 1c 2d 0e 1f 0g 2h
        l.add("i", pclass=1) # 0b 1c 2d 0e 1f 0g 2h 1i
        l.add("j", pclass=1) # 0b 2d 0e 1f 0g 2h 1i 1j
        l.add("k", pclass=0) # 2d 0e 1f 0g 2h 1i 1j 0k
        self.failUnlessEqual(len(l.getItems(None)), 8)
        self.failUnlessEqual(len(l.getItems(0)), 3)
        self.failUnlessEqual(len(l.getItems(1)), 3)
        self.failUnlessEqual(len(l.getItems(2)), 2)
        self.failUnlessEqual(l.getMessage(-1, None), ("k",))
        self.failUnlessEqual(l.getMessage(-1, 0), ("k",))
        self.failUnlessEqual(l.getMessage(-1, 1), ("j",))
        self.failUnlessEqual(l.getMessage(-1, 2), ("h",))
        self.failUnlessEqual(l.getMessage(0, None), ("d",))
        self.failUnlessEqual(l.getMessage(0, 0), ("e",))
        self.failUnlessEqual(l.getMessage(0, 1), ("f",))
        self.failUnlessEqual(l.getMessage(0, 2), ("d",))

    def testTree(self):
        l = log.Log("foo")
        # a
        #  b
        #   c
        #   d
        #  e
        #   f
        # g

        a = l.add("a")
        b = l.add("b", parent=a)
        l.add("c", parent=b)
        l.add("d", parent=b)
        e = l.add("e", parent=a)
        l.add("f", parent=e)
        l.add("g")

        items = l.getItems()
        self.failUnlessEqual(len(items), 2)
        A = items[0]
        self.failUnlessEqual(A.getMessage(), ("a",))
        Akids = A.getChildren()
        self.failUnlessEqual(len(Akids), 2)
        B,C = Akids
        self.failUnlessEqual(B.getMessage(), ("b",))
        Bkids = B.getChildren()
        self.failUnlessEqual(len(Bkids), 2)
        self.failUnlessEqual(Bkids[1].getMessage(), ("d",))

    def testObservers(self):
        l = log.Log("foo")
        events = []
        def observe(lognode, pclass, parent, events=events):
            events.append(lognode.getMessage())
        l.addObserver(observe)

        self.failIf(events)
        l.add("a")
        self.failUnlessEqual(events, [("a",)])
        l.add("b")
        self.failUnlessEqual(events, [("a",), ("b",)])
