
from twisted.trial import unittest
from twisted.internet import reactor
dr = unittest.deferredResult

import os
from petmail import mm_util, transport


class TestMixminion(unittest.TestCase):
    mm_path = None # set by the module-level setup() function

    def getMM(self):
        mm_path = os.environ.get("PETMAIL_TEST_MIXMINION", "mixminion")
        d = mm_util.getVersion(mm_path)
        version, error = dr(d)
        if not version:
            msg = "couldn't locate mixminion. Install it on your PATH " + \
                  "or define PETMAIL_TEST_MIXMINION=/path/to/mixminion"
            raise unittest.SkipTest(msg)
        return mm_path

    def testMint(self):
        mm_path = self.getMM()
        mmc = mm_util.MixminionContext("test_mm")
        mmc.set_mm_path(mm_path)

        surb1 = dr(mmc.mint("dest@example.com", "id1"))
        self.failUnless(surb1)
        self.failIf(surb1.granted)
        self.failIf(surb1.redeemed)
        self.failIf(surb1.expired)
        self.failUnlessEqual(surb1.identity, "id1")

        text = surb1.get()
        self.failUnless(surb1.granted)
        self.failUnless("BEGIN TYPE III REPLY BLOCK" in text)

        surb2 = dr(mmc.mint("dest@example.com", "id1"))
        self.failUnlessEqual(surb1.identity, surb2.identity)
        #self.failIfEqual(surb1.surbid, surb2.surbid) # need MM API first

    def testExpire(self):
        mm_path = self.getMM()
        mmc = mm_util.MixminionContext("test_mm")
        mmc.set_mm_path(mm_path)

        surb1 = dr(mmc.mint("dest@example.com", "id1"))
        results = []
        def expired(s, results=results):
            results.append(s)
        surb1.notifyOnExpiration(expired)
        self.failIf(results)
        mmc.expire(surb1)
        self.failUnless(surb1.expired)
        self.failUnlessEqual(results, [surb1])

    def receiveMessage(self, msg, source):
        self.messages.append((msg, source))

    def testDecodePlaintext(self):
        # this one actually spawns mixminion to unwrap 'plaintext' messages
        mm_path = self.getMM()
        self.mmc = mmc = mm_util.MixminionContext("test_mm")
        mmc.set_mm_path(mm_path)

        message = "message\x00\x01\x02" * 40
        encoded = transport.encapsulateInText(message)
        
        mmencoded = "\n" + encoded
        mmencoded = mmencoded.replace("\n--", "\n- --")
        mmencoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                    "Message-type: plaintext\n" + \
                    mmencoded + \
                    "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = transport.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return True
        i.canDecryptType3 = canDecryptType3
        def decryptType3(encoded, self=self):
            return self.mmc.decrypt(encoded)
        i.decryptType3 = decryptType3

        self.messages = []
        dr(i.gotText(mmencoded))
        self.failUnless(len(self.messages), 1)
        out, (inlet, idstring, surbid) = self.messages[0]

        self.failUnlessEqual(out, message)
        self.failUnlessEqual(inlet, i)
        self.failIf(idstring)
        self.failIf(surbid)

        # mixminion 0.0.6.2 adds its version/warning to stdout, making it
        # hard to separate from the decoded message. That will be fixed in
        # the next release.

        #self.failUnlessEqual(out, encoded)
        #self.failUnless(out.endswith(encoded))


    def testDecodeEncrypted(self):
        raise unittest.SkipTest("not yet implemented")
        # this one should actually spawn mixminion to unwrap 'encrypted'
        # messages

        mm_path = self.getMM()
        self.mmc = mmc = mm_util.MixminionContext("test_mm")
        mmc.set_mm_path(mm_path)

        # need multiple things to do this:
        #  must create a valid encrypted message (make SURB, send-to-SURB)
        #  must get the encrypted message (i.e. before it gets sent out)
        #  must hand message to the same MM environment which made the SURB

        message = "message\x00\x01\x02" * 40
        encoded = transport.encapsulateInText(message)

        encoded = "\n" + encoded
        encoded = encoded.replace("\n--", "\n- --")
        # i.e. so this nonsense won't cut it
        encoded = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----\n" + \
                  "Message-type: encrypted\n" + \
                  "Decoding-handle: blahblahblah\n" + \
                  encoded + \
                  "-----END TYPE III ANONYMOUS MESSAGE-----\n"

        i = transport.SMTPEncapsulatedInlet()
        i.parent = self
        def canDecryptType3(): return True
        i.canDecryptType3 = canDecryptType3
        def decryptType3(encoded, self=self):
            return self.mmc.decrypt(encoded)
        i.decryptType3 = decryptType3

        # ...


def MM_usable(mm_path):
    d = mm_util.getVersion(mm_path)
    version, error = dr(d)
    if not version:
        return "couldn't locate mixminion. Install it on your PATH " + \
               "or define PETMAIL_TEST_MIXMINION=/path/to/mixminion"
    mmc = mm_util.MixminionContext("test_mm")
    mmc.set_mm_path(mm_path)
    r = dr(mmc.checkDirectoryDownload())
    if not r:
        return "mixminion cannot contact its directory servers"
    return None

def setup():
    mm_path = os.environ.get("PETMAIL_TEST_MIXMINION", "mixminion")
    why = MM_usable(mm_path)
    if why:
        TestMixminion.skip = why
    else:
        TestMixminion.mm_path = mm_path

setup()
