
"""
This file contains various forms of Transport objects, which are used to get
an encapsulated message body (a binary string) to their receiver. The most
likely choice is a MIME-encoded rfc822 message send via the legacy SMTP
infrastructure, but anything which reliably gets the blob to the other end
will work.

The Transport is also responsible for doing an end-to-end acknowledgement,
retransmitting if necessary.
"""

import os, os.path, re
import email
from email import Encoders, Parser, Message
import base64

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

from twisted.python import log
from twisted.python.components import implements
from twisted.spread import pb
from twisted.cred import portal
from twisted.application import service, internet
from twisted.internet import protocol, error, defer, reactor

from twisted.protocols.smtp import sendmail
from twisted.protocols.pop3 import POP3Client
from twisted.protocols import imap4
from twisted.mail import relaymanager

from petmail import pbutil, interfaces, mm_util, util

# the recipient adds an Inlet to their agent
# the Inlet produces a transport string (.encode)
# the transport string is published in the ID Record
# the sender gets the ID Record and extracts the transport string
# the transport string has a prefix and the args
# the prefix is looked up in knownTransports and used to create a Transport

knownTransports = {}
def registerTransport(prefix, factory):
    knownTransports[prefix] = factory

def transportFromString(string, sender):
    if not string:
        return Null(None, sender)
    colon = string.index(":")
    type = string[:colon]
    args = string[colon+1:]
    transport = knownTransports[type](args, sender)
    transport.string = string # TODO? why?
    return transport

class BaseInlet(service.MultiService):
    __implements__ = (interfaces.IInlet, service.MultiService.__implements__)

    def setServiceParent(self, parent):
        assert implements(parent, interfaces.IAgent)
        service.MultiService.setServiceParent(self, parent)
    def encode(self):
        return self.description
    def gotMessage(self, encoded, source=(None, None)):
        self.parent.receiveMessage(encoded, (self, source[0], source[1]))
    def canDecryptType3(self):
        return self.parent.canDecryptType3()
    def decryptType3(self, mmencrypted):
        return self.parent.decryptType3(mmencrypted)

dummyInlets = {}
class DummyInlet(BaseInlet):
    """DummyInlet/DummyTransport are paired like PTY devices. Both live in
    the same process, and writing to one causes the message to be emitted on
    the other. Used only for testing.

    Dummy:name
    """

    def __init__(self, name):
        BaseInlet.__init__(self)
        self.name = name
        dummyInlets[self.name] = self
        self.description = "Dummy:%s" % self.name
    def accept(self, encrypted):
        return self.gotMessage(encrypted)
    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp(self.name, other.name))
    def __hash__(self):
        return hash((self.__class__, self.name))

class DummyTransport:
    __implements__ = interfaces.IUntraceableTransport,  # entirely local
    def __init__(self, args, sender):
        self.name = args
    def send(self, encapsulated):
        inlet = dummyInlets[self.name]
        return inlet.accept(encapsulated)
registerTransport("Dummy", DummyTransport)


class Null:
    __implements__ = interfaces.IUntraceableTransport,
    # Never Say Anything. Best kind of security in the world.

    def __init__(self, args, sender):
        pass
    def send(self, encapsulated):
        print "discarding message"
registerTransport("Null", Null)



class PBInlet(BaseInlet, pb.Root):
    __implements__ = (BaseInlet.__implements__, pb.Root.__implements__)

    def __init__(self, hostname, portnum):
        BaseInlet.__init__(self)
        self.description = "PB:%s:%d" % (hostname, portnum)
        self.hostname = hostname
        self.portnum = int(portnum)
        server = pbutil.listen(self, int(portnum))
        server.setServiceParent(self)

    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp((self.hostname, self.portnum),
                    (other.hostname, other.portnum)))
    def __hash__(self):
        return hash((self.__class__, self.hostname, self.portnum))

    def remote_message(self, encrypted):
        return self.gotMessage(encrypted)
        
    def startService(self):
        BaseInlet.startService(self)

    def stopService(self):
        BaseInlet.stopService(self)
        # probably need to wait a moment here if the socket is going to
        # be re-opened right away


class PBTransport:
    __implements__ = interfaces.ITransport,

    def __init__(self, args, sender):
        host, portnum = args.split(":")
        self.host = host
        self.portnum = int(portnum)
    def send(self, encapsulated):
        d = pbutil.connect(self.host, self.portnum)
        d.addCallback(self._connected, encapsulated)
        d.addErrback(log.err)
        return d

    def _connected(self, ref, encapsulated):
        d = ref.callRemote("message", encapsulated)
        d.addCallback(self._sent, ref)
        d.addErrback(log.err)
        return d
    def _sent(self, result, ref):
        ref.broker.transport.loseConnection()
    def __repr__(self):
        return "petmail.transport.PB<%s:%d>" % (self.host, self.portnum)
registerTransport("PB", PBTransport)


SMTP_SUBJECT = "PETMAIL MESSAGE"
SMTP_HEADER = "-----BEGIN PETMAIL MESSAGE-----"
SMTP_FOOTER = "-----END PETMAIL MESSAGE-----"
TYPE3_HEADER = "-----BEGIN TYPE III ANONYMOUS MESSAGE-----"
TYPE3_FOOTER = "-----END TYPE III ANONYMOUS MESSAGE-----"

class SMTPEncapsulatedInlet(BaseInlet):

    # the next three methods may raise an exception, in which case the error
    # should be logged and the message dropped (or perhaps put in a log
    # directory for a limited period of time)

    def gotText(self, text):
        """This method should be called when the inlet has received an
        incoming message ready to be scanned for the ---BEGIN PETMAIL.. line
        This particular function should be given the message as a string."""
        f = StringIO(text)
        return self.gotFile(f)

    def gotFile(self, fp):
        """This function receives the message as an open file. The object
        must support the readline() and seek() methods.
        """
        # the message is either a --PETMAIL MESSAGE-- or a --TYPE III--
        d = self.unwrapPadding(fp, acceptType3=True)
        d.addCallback(self._gotFile2)
        return d

    def _gotFile2(self, results):
        encoded, identity, surbid = results
        return self.gotMessage(encoded, (identity, surbid))

    def unwrapPadding(self, fp, acceptType3):
        """This method receives the message as an open file. The object
        must support the readline() and seek() methods. It returns 
        """
        mm = "Malformed message "
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError(mm + "(no BEGIN banner)")
            if line.rstrip() == SMTP_HEADER:
                line = fp.readline()
                if not line:
                    raise ValueError(mm + "(missing separator)")
                if line.strip():
                    raise ValueError(mm + "(unexpected headers)")
                break
            if line.rstrip() == TYPE3_HEADER:
                if acceptType3:
                    return self.unwrapType3(fp)
                else:
                    raise ValueError(mm + "(recursive type3)")
        body = ""
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError(mm + "(no END banner)")
            if line.rstrip() == SMTP_FOOTER:
                break
            body += line

        encoded = base64.decodestring(body)
        results = (encoded, None, None)
        return defer.succeed(results)


    def unwrapType3(self, fp):
        # first off, do we have the mixminion client? If so, use it to
        # unwrap the type3-encoding.
        if self.canDecryptType3():
            fp.seek(0)
            text = fp.read()
            d = self.decryptType3(text)
            d.addCallback(self._decryptedType3)
            return d
        # we've just read the BEGIN line. Next are the header lines
        plaintext = False
        while 1:
            line = fp.readline()
            if not line:
                raise ValueError("Malformed message")
            line = line.rstrip()
            if re.search(r'^Message-type:\s*plaintext$', line, re.I):
                plaintext = True
            elif re.search(r'^Message-type:', line, re.I):
                plaintext = False
            elif line == '':
                # end of headers
                break
        if not plaintext:
            msg = "Type3 message requires mixminion executable"
            raise interfaces.UndecryptableError(msg)
        body = ""
        while 1:
            line = fp.readline()
            if not line:
                # EOF too early
                raise ValueError("Malformed message")
            if line.rstrip() == TYPE3_FOOTER:
                break
            if line.startswith("- "):
                line = line[2:]
            body += line
        return self.unwrapPadding(StringIO(body), acceptType3=False)

    def _decryptedType3(self, results):
        text, idstring, surbid = results
        # unwrap the plaintext, but tack on the idstring/surbid
        d = self.unwrapPadding(StringIO(text), acceptType3=False)
        d.addCallback(self._unwrappedDecryptedType3, idstring, surbid)
        return d

    def _unwrappedDecryptedType3(self, results, idstring, surbid):
        text, dummy1, dummy2 = results
        return (text, idstring, surbid)


class PollingSMTPInlet(SMTPEncapsulatedInlet):
    startedTimer = False
    pollinterval = 600 # seconds

    def startService(self):
        if not self.startedTimer:
            # note that the TimerService is not pickleable until it has been
            # stopped.
            # TODO: consider using the deferred to prevent overlapping checks
            timer = internet.TimerService(self.pollinterval, self.check)
            timer.setServiceParent(self)
            self.startedTimer = True
        SMTPEncapsulatedInlet.startService(self)

    def check(self):
        """I check the mailbox and return a Deferred which will fire when the
        check has completed.
        """
        raise Unimplemented

class MaildirInlet(PollingSMTPInlet):
    """I watch a maildir for inbound messages."""

    pollinterval = 10 # seconds

    def __init__(self, basedir, address):
        PollingSMTPInlet.__init__(self)
        self.basedir = basedir
        self.address = address
        self.newdir = os.path.join(os.path.expanduser(basedir), "new")
        self.olddir = os.path.join(os.path.expanduser(basedir), "cur")
        self.description = "SMTP:%s" % address

    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp((self.basedir, self.address),
                    (other.basedir, other.address)))
    def __hash__(self):
        return hash((self.__class__, self.basedir, self.address))

    def check(self):
        for f in os.listdir(self.newdir):
            newfile = os.path.join(self.newdir, f)
            oldfile = os.path.join(self.olddir, f)
            self.gotFile(open(newfile, "r"))
            os.rename(newfile, oldfile)
                      

class MailSocketProtocol(protocol.Protocol):
    def __init__(self):
        self.data = ""
    def dataReceived(self, data):
        self.data += data
    def connectionLost(self, reason):
        if reason.check(error.ConnectionDone):
            self.factory.parent.gotText(self.data)

class MailSocket(SMTPEncapsulatedInlet):
    # [MailSocket:]9990:alice@foo.com
    # socketnum : inletaddress

    def __init__(self, portnum, address):
        SMTPEncapsulatedInlet.__init__(self)
        assert type(portnum) == int
        self.portnum = portnum
        self.address = address
        f = protocol.ServerFactory()
        f.protocol = MailSocketProtocol
        f.parent = self
        server = internet.TCPServer(portnum, f)
        server.setServiceParent(self)
        self.description = "SMTP:%s" % address

    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp((self.portnum, self.address),
                    (other.portnum, other.address)))
    def __hash__(self):
        return hash((self.__class__, self.portnum, self.address))

# IMAP grabber

class IMAPPollProtocol(imap4.IMAP4Client):
    def serverGreeting(self, caps):
        # caps is a dictionary. caps['AUTH'] is a list of login methods
        d = self.login(self.factory.username, self.factory.password)
        d.addCallback(self.loggedIn)
        d.addCallback(self.factory.complete)
        d.addErrback(self.factory.failed)

    def loggedIn(self, *args):
        if self.factory.doDelete:
            d = self.select(self.factory.folder) # read-write
        else:
            d = self.examine(self.factory.folder)
        d.addCallback(self.examine_cb)
        return d

    def examine_cb(self, info):
        if info['EXISTS']:
            d = self.fetchHeaders("1:*")
            d.addCallback(self.gotHeader)
            return d
        else:
            return self.done()

    def gotHeader(self, msgs):
        ids = msgs.keys()
        ids.sort()
        wanted = imap4.MessageSet()
        for msgid in ids:
            headertext = msgs[msgid]['RFC822.HEADER']
            m = email.message_from_string(headertext)
            print "%s: %s" % (msgid, m['subject'])
            if m['subject'] == "PETMAIL MESSAGE":
                wanted.add(msgid)
        print "wanted:", wanted
        if wanted:
            d = self.fetchBody(wanted)
            d.addCallback(self.gotBody, wanted)
            return d
        else:
            return self.done()

    def gotBody(self, msgs, wanted):
        print "gotBody"
        ids = msgs.keys()
        ids.sort()
        for msgid in ids:
            text = msgs[msgid]['RFC822.TEXT']
            self.factory.inlet.gotText(text)
        if self.factory.doDelete:
            print "deleting:", wanted
            # TODO: twisted-1.1.2 will let us remove the str()
            d = self.setFlags(str(wanted), ("\\Deleted",))
            d.addCallback(self.didDelete)
            return d
        else:
            return self.done()

    def didDelete(self, *args):
        print "didDelete", args
        return self.done()

    def done(self):
        d = self.close()
        d.addCallback(self._closed)
        return d

    def _closed(self, results):
        self.transport.loseConnection()

class IMAPPollerFactory(protocol.ClientFactory):
    protocol = IMAPPollProtocol

    def __init__(self, username, password, folder):
        self.username = username
        self.password = password
        self.folder = folder # might be an empty string
        self.deferred = defer.Deferred()
        self.done = False

    def clientConnectionFailed(self, connector, reason):
        self.failed(reason)

    def clientConnectionLost(self, connector, reason):
        self.failed(reason)

    def complete(self, results):
        print "complete"
        self.done = True
        self.deferred.callback(results)

    def failed(self, why):
        if not self.done:
            self.done = True
            print "failure", why
            self.deferred.errback(why)

class IMAPPoller(PollingSMTPInlet):
    doDelete = True

    def __init__(self, host, portnum, username, password, pollinterval,
                 folder, address):
        PollingSMTPInlet.__init__(self)
        self.host = host
        self.portnum = portnum
        self.username = username
        self.password = password
        self.pollinterval = pollinterval
        self.address = address
        self.folder = folder # might be an empty string
        self.description = "SMTP:%s" % address

    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp((self.host, self.portnum, self.username,
                     self.password, self.pollinterval, self.folder,
                     self.address),
                    (other.host, other.portnum, other.username,
                     other.password, other.pollinterval, other.folder,
                     other.address)))

    def __hash__(self):
        return hash((self.__class__, self.host, self.portnum, self.username,
                     self.password, self.pollinterval, self.folder,
                     self.address))

    def check(self):
        print "connecting.."
        f = IMAPPollerFactory(self.username, self.password, self.folder)
        f.doDelete = self.doDelete
        f.inlet = self
        reactor.connectTCP(self.host, self.portnum, f)
        return f.deferred



class POPPollProtocol(POP3Client):
    # incomplete
    def handle_WELCOME(self, line):
        POP3Client.handle_WELCOME(self, line)
        self.user(self.factory.poller.username)
    def handle_USER(self, data):
        pass

    
class POPPoller(PollingSMTPInlet):

    def __init__(self, host, portnum, username, password, pollinterval,
                 address):
        raise Unimplemented # incomplete
        SMTPEncapsulatedInlet.__init__(self)
        self.host = host
        self.portnum = portnum
        self.username = username
        self.password = password
        self.pollinterval = pollinterval
        self.address = address
        self.description = "SMTP:%s" % address

    def __cmp__(self, other):
        return (cmp(self.__class__, other.__class__) or
                cmp((self.host, self.portnum, self.username,
                     self.password, self.pollinterval, self.address),
                    (other.host, other.portnum, other.username,
                     other.password, other.pollinterval, other.address)))

    def __hash__(self):
        return hash((self.__class__, self.host, self.portnum, self.username,
                     self.password, self.pollinterval, self.address))

    def check(self):
        print "checking"
                      


# SMTP-encapsulating output. The Agent has a .smtpScheme string which defines
# a preference for how we should do outbound SMTP. This scheme is
# "Mail:relay:fromaddr" for simple mailing, and "Mixminion:??" for
# send-through-mixminion

smtpSchemes = {}
def makeSMTPTransport(args, sender):
    dest = args
    secure = sender.pseudonymous
    if secure:
        return MailAnonymously(dest)
    else:
        scheme, smarthost, fromEmail = sender.parent.smtpScheme.split(":")
        # TODO: fromEmail = sender.fromSMTPAddr
        return JustMailIt(dest, fromEmail, smarthost)
registerTransport("SMTP", makeSMTPTransport)

def encapsulateInText(encrypted):
    """This puts the PETmail message body into a ---PETMAIL MESSAGE---
    payload block. It returns a string ready to be inserted into the body
    of a Message object. The subject line of the resulting message must
    be 'PETMAIL MESSAGE'."""

    return util.armor(encrypted, "MESSAGE")

class JustMailIt(object):
    """This Transport class uses twisted.mail to deliver the message via SMTP
    to a designated relay host. For normally-abled unix boxes that run an MTA
    like sendmail, this defaults to 'localhost'. Others will need to set it
    to a host which *is* capable of sending mail.
    """
    __implements__ = interfaces.ITransport,

    def __init__(self, dest, fromEmail, smarthost="localhost"):
        self.dest = dest
        self.fromEmail = fromEmail
        self.smarthost = smarthost

    def send(self, encrypted):
        """This returns a Deferred which will fire when the message has been
        sent successfully."""
        m = Message.Message()
        m.set_payload(encapsulateInText(encrypted))
        m['Subject'] = SMTP_SUBJECT
        m['X-PETmail-URL'] = "http://petmail.lothar.com/"
        m['From'] = self.fromEmail
        m['To'] = self.dest
        s = m.as_string()
        return sendmail(self.smarthost, self.fromEmail, self.dest, s)


class MailAnonymously(object):
    """This subclass sends the message out through a remailer network to a
    known recipient, making it nigh-impossible to figure out where it came
    from. The only information available to the recipient (or an
    eavesdropper) is the contents of the IDRecord.

    This is the anonymous-sender, non-anonymous-recipient case. For
    bidirectional anonymity or anonymous-recipient, use MailToSURB."""
    __implements__ = interfaces.IUntraceableTransport,

    def __init__(self, dest):
        self.dest = dest

    def send(self, encrypted):
        body = encapsulateInText(encrypted)
        cmd = ("mixminion", "send",
               "--to", self.dest,
               "--subject", SMTP_SUBJECT,
               "--queue",
               )
        pp = mm_util.HelperPP({0: body}, [1,2])
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r"})
        d.addCallback(self._sent)
        d.addErrback(self._sendError)
        return d

    def _sent(self, outputs):
        return True

    def _sendError(self, results):
        why, outputs = results
        print "mixminion send failed"
        print why
        print "stdout:", results[1]
        print "stderr:", results[2]
        return interfaces.Unsendable("mixminion send failed")


class MailToSURB(object):
    """This subclass sends the message out through a remailer network to an
    anonymous recipient.

    This is the bidirectional anonymity / anonymous-recipient case."""
    __implements__ = interfaces.IUntraceableTransport,

    def send(send, encrypted):
        body = encapsulateInText(encrypted)
        cmd = ("mixminion", "send",
               "--reply-block", self.destSURB,
               "--subject", SMTP_SUBJECT,
               "--queue",
               )
        pp = mm_util.HelperPP({0: body}, [1,2])
        d = pp.deferred
        p = reactor.spawnProcess(pp, cmd[0], cmd, env=None, path=None,
                                 childFDs={0:"w", 1:"r", 2:"r"})
        d.addCallback(self._sent)
        d.addErrback(self._sendError)
        return d

    def _sent(self, outputs):
        return True

    def _sendError(self, results):
        why, outputs = results
        print "mixminion send failed"
        print why
        print "stdout:", results[1]
        print "stderr:", results[2]


smtpSchemes["MixminionSURB"] = MailToSURB

class UsenetDropbox(object):

    """This subclass posts the message to a Usenet group, probably
    alt.anonymous.messages, where the recipient can pick it up at their
    (fairly anonymous) leisure. The sender is not particularly anonymous
    unless they send the message through a remailer network."""

    pass

# TODO: the SMTP Tranports have two kinds of configurable state: that which
# comes from the sender's configuration, and that which comes from the
# recipient's TransportRecord.

# recipient defines:
#  "I want messages via SMTP at address bob@foo.com"
#  "I want messages to be posted to alt.anonymous.messages with subject
#   '123'"

# sender defines:
#  "All SMTP messages should be sent with a From: address of alice@foo.com"
#  "All SMTP messages should be sent through smarthost smtp.isp.com"
#  "All SMTP messages should be sent through mixminion"
#  "Usenet posts should be posted to nntp.isp.com"
#  "Usenet posts should be mail-to-newsed through mixminion and m2n.blah.com"

# Likewise, the Inlet contains state that is purely local, which should not
# be published in their TransportRecord. The published data is listed above.

# recipient-local data:
#  "I get SMTP mail by sucking POP from user/pass @pop.isp.com every 5min"
#  " or by IMAP"
#  "I get SMTP by listening on port 25"
#  "I get SMTP by sucking on a maildir at ~/Maildir-petmail every 5 minutes"
#  "I get SMTP by watching ~/Maildir-petmail for changes with FAM or such"
#  "I get SMTP by one of the above, and sometimes doing nymserver-IMAP
#    polling when the nymserver tells me there are a lot of messages queued,
#    and oh by the way please keep the nymserver supplied with SURBs"
#  "I get Usenet posts by polling nntp.isp.com every 30 minutes"
#  "I get SMTP by listening on localhost:9990 or the ~/.petmail-inlet UNIX
#    socket, because my .forward/.qmail says '|nc localhost 9990'"
